from . import redis
from . import local

CACHE_PKG=local
T_CACHE_TYPE=["local", "redis"]
CACHE_SYMBOL="local"

def setCacheType(_type, **kwargs):
    if _type not in T_CACHE_TYPE:
        raise ValueError(f"{_type} is not registred cache type {T_CACHE_TYPE}")
    global CACHE_PKG, CACHE_SYMBOL
    CACHE_PKG = redis if _type == "redis" else local
    CACHE_SYMBOL = _type
    print(f"Set cache to {_type}")

    if _type == 'redis':
        p = {}
        if 'rp' in kwargs:
            p['port'] = kwargs['rp']
        if 'rh' in kwargs:
            p['host'] = kwargs['rh']
        
        CACHE_PKG.setDatabaseParameters(**p)

def storeTreeByTaxid(tree, taxid):
    return CACHE_PKG.storeTreeByTaxid(tree, taxid)

def delTreeByTaxids(taxids):
    return CACHE_PKG.delTreeByTaxids(taxids)

def getTaxidKeys():
    return CACHE_PKG.getTaxidKeys()

def getUniversalTree(taxid, raw=False):
    # if not local deserialize or deeper

    return CACHE_PKG.getUniversalTree(taxid, raw=raw)

def getUniversalVector(taxid):
    try:
        vec = CACHE_PKG.getUniversalVector(taxid)
        print(f"{taxid} Universal vector in cache")
    except KeyError:
        try:
            tree = CACHE_PKG.getUniversalTree(taxid)
        except KeyError:
            raise KeyError(f"Vector error, No taxid {taxid} in stores")

        print(f"Building {taxid} Universal vector")
        vec = tree.vectorize()
        CACHE_PKG.storeVectorByTaxid(vec, taxid)
    
    return vec

def clear():
    CACHE_PKG.clear()

def status():
    if CACHE_SYMBOL == 'redis':
        return len(listTrees()), len(listVectors()), len(listCulled())

def listTrees():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listTreeKey(prefix=False) ]

def listVectors():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listVectorKey(prefix=False) ]
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL KEYS ITER")

def listCulled():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listCulledVectorKey(prefix=False) ]
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL KEYS ITER")


def getCulledVector(taxid, cmin, cmax, fmax):
    if CACHE_SYMBOL == 'redis':
        _ = CACHE_PKG.getCulledVector(taxid, cmin, cmax, fmax)
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL getCulledVector")
    return _

def storeCulledVector(vector, taxid, cmin, cmax, fmax):
    if CACHE_SYMBOL == 'redis':
        _ = CACHE_PKG.storeCulledVector(vector, taxid, cmin, cmax, fmax)
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL storeCulledVector")
    return _   

def listMissUniversalVector():
    _treeID   = set( listTrees() )
    _vectorID = set( listVectors() )
    return list(_treeID - _vectorID)

def buildUniversalVector():
    print("Running unBuildtUniversalVectorIter")
    _ = listMissUniversalVector()
    print(f"Set of trees to vectorize {_}")
    for bKey in _:
        print(f"-->{type(bKey)}")
        print(f"Build vector for {bKey}")
        tree = CACHE_PKG.getUniversalTree(bKey)
        CACHE_PKG.storeVectorByTaxid(tree, bKey)

"""
def storeTreeByTaxid(tree, taxid):
    return FN_HANDLER['storeTreeByTaxid'](tree, taxid)


def taxidKeys():
    return FN_HANDLER['getTaxidKeys']

def getUniversalTree(taxid):
    # if not local deserialize or deeper

    return FN_HANDLER['getUniversalTree'](taxid)

# Can be called for construction of vector banking
def getUniversalVector(taxid):
    try:
        vec = FN_HANDLER['getUniversalVector'](taxid)
        print(f"Found in vector cache {taxid}")
    except KeyError:
        try:
            tree = FN_HANDLER['getUniversalTree'](taxid)
        except KeyError:
            raise KeyError(f"Vector errro, No taxid {taxid} in stores")

        print(f"Creating vectors for {taxid}")
        vec = tree.vectorize()
        FN_HANDLER['storeVectorByTaxid'](vec, taxid)
    
    return vec
"""


UNIVERSAL_TREES = {}

UNIVERSAL_VECTORS = {}

def clear():
    global UNIVERSAL_TREES, UNIVERSAL_VECTORS
    print(f"Clearing local stores content")

    UNIVERSAL_VECTORS = {}
    UNIVERSAL_TREES   = {}

def delTreeByTaxids(*taxids):
    miss = []
    for taxid in taxids:
        if not taxid in UNIVERSAL_TREES:
            miss.append(taxid)
        else:
            UNIVERSAL_TREES.pop(taxid)
    if miss:
        raise KeyError(f"{miss} to delete elements not found in local store")

def storeTreeByTaxid(tree, taxid):
    print(f"localTreeStoring -->{taxid}")

    global UNIVERSAL_TREES
    if taxid in UNIVERSAL_TREES:
        raise KeyError(f"{taxid} tree already exists in local store")
        
    UNIVERSAL_TREES[taxid] = tree

def getUniversalTree(taxid):
    if not taxid in UNIVERSAL_TREES:
        raise KeyError(f"No taxid {taxid} found in local store")
    
    return UNIVERSAL_TREES[taxid]

def storeVectorByTaxid(vector, taxid):
    print(f"localVectorStoring -->{taxid}")

    global UNIVERSAL_VECTORS
    if taxid in UNIVERSAL_VECTORS:
        raise KeyError(f"{taxid} vector already exists in local store")
        
    UNIVERSAL_VECTORS[taxid] = tree

def getUniversalVector(taxid):
    if not taxid in UNIVERSAL_VECTORS:
        raise KeyError(f"No taxid {taxid} found in vector local store")
    
    return UNIVERSAL_VECTORS[taxid]

def getTaxidKeys():
    return list(UNIVERSAL_TREES.keys())

def listTreeKey(*args,**kwargs):
    return list(UNIVERSAL_TREES.keys())

def listVectorKey(*args,**kwargs):
    return list(UNIVERSAL_VECTORS.keys())



from pyrediscore import connect, store, get, delete, listKey, setDatabaseParameters
from ...data_objects import loadUnivGO

@connect
@store
def storeTreeByTaxid(tree, taxid, *args, **kwargs):
    return  f"tree:{taxid}", tree.serialize()

@connect
@store
def storeVectorByTaxid(tree, taxid, *args, **kwargs):
    return  f"vector:{taxid}", tree.vectorize()

@connect
@store
def storeCulledVector(vector, taxid, cmin, cmax, fmax, *args, **kwargs):
    return  f"_vector:{taxid}:{cmin}:{cmax}:{fmax}", vector
    
@connect
@delete
def delTreeByTaxids(taxids, *args, **kwargs):
    return [f"tree:{_}" for _ in taxids]

@connect
@delete
def delVectorByTaxids(taxids, *args, **kwargs):
    return [f"vector:{_}" for _ in taxids]

@connect
@get
def getUniversalTree(taxid, *args, raw=False, **kwargs):
    return (f"tree:{taxid}", loadUnivGO)

@connect
@get
def getUniversalVector(taxid, *args, raw=False, **kwargs):
    return (f"vector:{taxid}", None)

@connect
@get
def getCulledVector(taxid, cmin, cmax, fmax, *args, raw=False, **kwargs):
    return  (f"_vector:{taxid}:{cmin}:{cmax}:{fmax}", None)

@connect
@listKey
def listTreeKey(*args, prefix=False, **kwargs):
    return ('tree:*', 'tree:')

@connect
@listKey
def listVectorKey(*args, prefix=False, **kwargs):
    return ('vector:*', 'vector:')

@connect
@listKey
def listCulledVectorKey(*args, prefix=False, **kwargs):
    return ('_vector:*', '_vector:')

def clearVectors(*args, **kwargs):
    keyList = [ _ for _ in listVectorKey() ]
    print(f"Clearing following vector elements content:\n\t{keyList}")
    delVectorByTaxids(keyList)
    
def clearTrees(*args, **kwargs):
    keyList = [ _ for _ in listTreeKey() ]
    print(f"Clearing following tree elements content:\n\t{keyList}")
    delTreeByTaxids(keyList)
    
def clear(*args, **kwargs):
    clearVectors(*args, **kwargs)
    clearTrees(*args, **kwargs)

"""
scan 0 MATCH *11*
"""

"""
def delTreeByTaxids(taxids):
    miss = []
    print(f"delete redis {taxids}")
    r = redis.Redis(host=HOST, port=PORT, db=0)       
    for taxid in taxids:
        _ = r.delete(f"tree:{taxid}")
        print(f"delete redis status {str(_)}")
        if int(str(_)) != 1:
            miss.append(taxid)
    return miss:
    
def storeTreeByTaxid(tree, taxid):
    print(f"redis taxid storage adding {taxid} {tree}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    if r.exists(f"tree:{taxid}"):
        raise KeyError(f"StoreTree error: taxid {taxid} already exists in store")
    d = tree.serialize()
    r.set(taxid, json.dumps(d))

def storeVectorByTaxid(vector, taxid):
    print(f"redis taxid storage adding {taxid} {vectors}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    if r.exists(f"vector:{taxid}"):
        raise KeyError(f"StoreVector error: taxid {taxid} already exists in store")
    
    r.set(taxid, json.dumps(vector))

def getUniversalTree(taxid, raw=False):
    print(f"redis taxid storage getting {taxid}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    _ = r.get(f"tree:{taxid}")
    if not _:
        raise KeyError(f"No taxid {taxid} found in tree store")
    # _ is bytes
    return _ \
        if raw \
        else loadUnivGO( json.loads(_) )

"""